from .circuit import Circuit, DeclareCoreirCircuit
from .array import Array
from .bits import Bits
from .conversions import array
from .t import Type, In, Out
from .tuple import Product
from .bitutils import clog2
from .generator import Generator
from .frontend.coreir_ import DefineCircuitFromGeneratorWrapper, GetCoreIRBackend
from .wire import wire
from .interface import IO


class CoreIRCommonLibMuxN(Generator):
    @staticmethod
    def generate(N: int, width: int):
        return DeclareCoreirCircuit(
            f"coreir_commonlib_mux{N}x{width}",
            *["I", In(Product.from_fields("anon",
                                          dict(data=Array[N, Bits[width]],
                                               sel=Bits[clog2(N)]))),
              "O", Out(Bits[width])],
            coreir_name="muxn",
            coreir_lib="commonlib",
            coreir_genargs={"width": width, "N": N}
        )


class Mux(Generator):
    @staticmethod
    def generate(height: int, T: Type):
        # TODO: Type must be hashable so we can cache
        N = T.flat_length()

        io_dict = {}
        for i in range(height):
            io_dict[f"I{i}"] = In(T)
        io_dict["S"] = In(Bits[clog2(height)])
        io_dict["O"] = Out(T)

        class Mux(Circuit):
            io = IO(**io_dict)

            mux = CoreIRCommonLibMuxN(height, N)
            data = []
            for i in range(height):
                data.append(Bits[N](getattr(io, f"I{i}").flatten()))
            wire(mux.I.data, array(data))
            wire(mux.I.sel, io.S)
            wire(mux.O, io.O)
        return Mux


def slice(value: Bits, start: Bits, width: int):
    """
    Dynamic slice based off the value of `start` with constant `width`

    Example:
    ```
    class TestSlice(m.Circuit):
        # Slice a 6 bit window of I using x as the start index
        IO = m.IO(
            I=m.In(m.Bits[10]),
            x=m.In(m.Bits[2]),
            O=m.Out(m.Bits[6])
        )

        io.O @= m.slice(io.I, start=io.x, width=6)
    ```

    Notes:
    * For now we only support slicing bits with bits index
    * We could add support for slicing Arrays
    * We could restrict index to be a UInt
    """
    # Construct an array where the index `i` is the slice of bits from `i` to
    # `i+width`, index into this array using `start`
    return array([value[i:i + width] for i in range(len(value) - width)])[start]



from abc import ABC, abstractmethod
from itertools import chain
from collections import OrderedDict
from .conversions import array
from .ref import InstRef, DefnRef, LazyDefnRef, NamedRef
from .t import Type, Kind, MagmaProtocolMeta, Direction
from .clock import Clock, ClockTypes
from .array import Array
from .tuple import Tuple
from .compatibility import IntegerTypes, StringTypes


__all__  = ['DeclareInterface']
__all__ += ['Interface']
__all__ += ['AnonymousInterface']
__all__ += ['InterfaceKind']
__all__ += ['DeclareLazyInterface']
__all__ += ['IO']


def _flatten(l):
    """
    Flat an iterable of iterables to list.
    """
    return list(chain(*l))


def _make_interface_name(decl):
    return f"Interface({', '.join([str(d) for d in decl])})"


def _is_valid_port(port):
    return isinstance(port, (Kind, Type, MagmaProtocolMeta))


def _parse(decl):
    """
    Parse argument declaration of the form:

        (name0, type0, name1, type1, ..., namen, typen)
    """
    if len(decl) % 2:
        raise ValueError(f"Expected even number of arguments, got {len(decl)}")

    names = decl[::2]
    ports = decl[1::2]
    # If name is empty, convert to the index.
    names = [name if name else i for i, name in enumerate(names)]
    # Check that all ports are given as instances of Kind or Type.
    if not all(_is_valid_port(port) for port in ports):
        raise ValueError(f"Expected kinds or types, got {ports}")

    return names, ports


def _make_interface_args(decl, renamed_ports, inst, defn):
    names, ports = _parse(decl)  # parse the class Interface declaration
    args = OrderedDict()
    for name, port in zip(names, ports):
        if   inst: ref = InstRef(inst, name)
        elif defn: ref = DefnRef(defn, name)
        else:      ref = NamedRef(name)

        if name in renamed_ports:
            ref.name = renamed_ports[name]
        if defn:
           port = port.flip()
        if isinstance(port, MagmaProtocolMeta):
            args[name] = port._from_magma_value_(port._to_magma_()(name=ref))
        else:
            args[name] = port(name=ref)

    return args


def _make_wire_str(driver, value, wired):
    """
    Emit the wiring string for `driver` and `value` for the `__repr__` of an
    interface.

    Used by `_make_wires`, `wired` is a set of previously wired values, so in
    cases of fan-out wires are not done more than once.

    Handles non-whole values (e.g. arrays/tuple constructed with values coming
    from multiple sources) by emitting the wiring of the child values.
    """
    if (value, driver) in wired:
        return ""
    wired.add((value, driver))
    if isinstance(driver, (Array, Tuple)) and \
            not driver.iswhole(driver.ts):
        return "".join(_make_wire_str(d, v, wired)
                       for d, v in zip(driver, value))
    iname = value.name.qualifiedname()
    oname = driver.name.qualifiedname()
    return f"wire({oname}, {iname})\n"


def _make_wires(value, wired):
    """
    Make the wires used in the `__repr__` of an interface.

    Handles the recursive types with mixed direction (e.g. Tuples containing
    inputs and outputs)

    Handles anonymous values by skipping them (we don't include anonymous
    temporaries in the representation)

    Handles non-whole values (e.g. arrays/tuple constructed with values coming
    from multiple sources)

    Traces from `value` to the source driver, emitting the wiring of the
    temporaries along the way.
    """
    s = ""
    if value.is_output():
        return s
    if isinstance(value, (Array, Tuple)) and \
            not value.is_input() and \
            not value.is_output() and \
            not value.is_inout():
        # Mixed
        for v in value:
            s += _make_wires(v, wired)
        return s
    driver = value.value()
    if driver is None:
        return s
    if isinstance(value, (Array, Tuple)) and \
            not driver.iswhole(driver.ts):
        for elem in value:
            s += _make_wires(elem, wired)
        return s
    while driver is not None and driver.name.anon():
        # Skip anon values
        driver = driver.value()
    while driver is not None:
        if (driver, value) in wired:
            break
        s += _make_wire_str(driver, value, wired)
        if not driver.is_output():
            value = driver
            driver = driver.value()
        else:
            driver = None
    return s


class _Interface(Type):
    """
    Abstract Base Class for an Interface.
    """
    def __str__(self):
        return str(type(self))

    def __repr__(self):
        s = ""
        wired = set()
        for name, value in self.ports.items():
            if value.is_output():
                continue
            s += _make_wires(value, wired)
        return s

    @classmethod
    def items(cls):
        return cls.ports.items()

    def __iter__(self):
        return iter(self.ports)

    def __len__(self):
        return len(self.ports.keys())

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.ports[key]
        if isinstance(key, int):
            return self.arguments()[key]
        if isinstance(key, slice):
            return array([self[i] for i in range(*key.indices(len(self)))])
        raise ValueError(f"Expected key as str, int, or slice, got {key} "
                         f"({type(key)})")

    def arguments(self):
        """Return all the argument ports."""
        return list(self.ports.values())

    def is_input(self, port, include_clocks=False):
        return port.is_input() and \
            (not isinstance(port, ClockTypes) or include_clocks)

    def inputs(self, include_clocks=False):
        """Return all the argument input ports."""
        return list(filter(lambda x: self.is_input(x, include_clocks),
                           self.ports.values()))

    def inputs_by_name(self, include_clocks=False):
        inputs = []
        for name, port in self.ports.items():
            if self.is_input(port):
                inputs.append(name)
        return inputs

    def outputs(self):
        """Return all the argument output ports."""
        return list(filter(lambda port: port.is_output(), self.ports.values()))

    def args(self):
        """Return all the arguments as name, port."""
        return _flatten(self.ports.items())

    def decl(self):
        """
        Return all the arguments as name, flip(port) (same as the declaration).
        """
        return _flatten([name, type(port).flip()]
                        for name, port in self.ports.items())

    def inputargs(self):
        """Return all the input arguments as name, port."""
        return _flatten([name, port] for name, port in self.ports.items()
                        if port.is_input() and not isinstance(port, ClockTypes))

    def outputargs(self):
        """Return all the output arguments as name, port."""
        return _flatten([name, port] for name, port in self.ports.items()
                        if port.is_output())

    def clockargs(self):
        """Return all the clock arguments as name, port."""
        return _flatten([name, port] for name, port in self.ports.items()
                        if isinstance(port, ClockTypes))

    def clockargnames(self):
        """Return all the clock argument names."""
        return [name for name, port in self.ports.items()
                if isinstance(port, ClockTypes)]

    def isclocked(self):
        """Return True if this interface has a Clock."""
        return any(isinstance(port, ClockType) for
                   port in self.ports.values())

class Interface(_Interface):
    """Interface class."""
    def __init__(self, decl, renamed_ports={}):
        """
        This function assumes the port instances are provided:
            e.g. Interface('I0', In(Bit)(), 'I1', In(Bit)(), 'O', Out(Bit)())
        """
        names, ports = _parse(decl)
        args = OrderedDict()
        for name, port in zip(names, ports):
            if isinstance(name, IntegerTypes):
                name = str(name)  # convert integer to str, e.g. 0 to "0"
            if name in renamed_ports:
                raise NotImplementedError("Port renaming not implemented")
            args[name] = port
        self.ports = args

    def __str__(self):
        s = ", ".join(f"{k}: {v}" for k, v in self.ports.items())
        return f"Interface({s})"


class AnonymousInterface(Interface):
    """
    Anonymous interfaces can have ports that are not inputs/output, so we need
    ot see if they trace to an input or if they are wired to inputs

    TODO: Need tests for all these cases
    """
    def outputs(self):
        return list(filter(
            lambda port: port.is_output() or port.trace() is None and
            not port.wired() and not port.is_input() and not port.is_inout(),
            self.ports.values()))

    def is_input(self, port, include_clocks=False):
        return port.is_input() or port.trace() is None and \
            port.wired() and not port.is_output() and not port.is_inout() and \
            (not isinstance(port, ClockTypes) or include_clocks)


class _DeclareInterface(_Interface):
    """
    _DeclareInterface class.

    First, an Interface is declared:
        Interface = DeclareInterface('I0', In(Bit), 'I1', In(Bit), 'O', Out(Bit))

    Then, the interface is instanced:
        interface = Interface()
    """
    def __init__(self, renamed_ports={}, inst=None, defn=None):
        self.ports = _make_interface_args(self.Decl, renamed_ports, inst, defn)

class _DeclareLazyInterface(_Interface):
    """_DeclareLazyInterface class"""
    def __init__(self, renamed_ports={}, inst=None, defn=None):
        # This interface declaration is only lazy for module definitions (not
        # instances). If @defn is not supplied, then we use the standard
        # (non-lazy) interface construction logic.
        if not defn:
            self.ports = _make_interface_args(self.Decl, renamed_ports, inst,
                                              defn)
            return
        self.io.bind(defn)  # bind IO to @defn
        args = OrderedDict()
        for name, port in self.io.ports.items():
            if name in renamed_ports:
                port.name.name = renamed_ports[name]
            args[name] = port

        self.ports = args


class InterfaceKind(Kind):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ports = []
        key = None
        for i, arg in enumerate(cls.Decl):
            if i % 2 == 0:
                key = arg
            else:
                ports.append((key, arg))
        cls.ports = OrderedDict(ports)

    def items(cls):
        return cls.ports.items()

    def __iter__(cls):
        return iter(cls.ports)

    def __str__(cls):
        args = []
        for i, arg in enumerate(cls.Decl):
            if i % 2 == 0:
                args.append(f"\"{arg}\"")
            else:
                args.append(str(arg))
        return ", ".join(args)

    def __eq__(cls, rhs):
        return cls.Decl == rhs.Decl

    __ne__=Kind.__ne__
    __hash__=Kind.__hash__


def DeclareInterface(*decl, **kwargs):
    """Interface factory function."""
    name = _make_interface_name(decl)
    dct = dict(Decl=decl, **kwargs)
    return InterfaceKind(name, (_DeclareInterface,), dct)


def DeclareLazyInterface(io, **kwargs):
    """LazyInterface factory function"""
    decl = io.decl()
    name = _make_interface_name(decl)
    dct = dict(io=io, Decl=io.decl(), **kwargs)
    return InterfaceKind(name, (_DeclareLazyInterface,), dct)


class IOInterface(ABC):
    """Abstract base class for IO-like classes"""
    @property
    @abstractmethod
    def ports(self):
        raise NotImplementedError()

    @abstractmethod
    def bind(self, defn):
        raise NotImplementedError()

    @abstractmethod
    def decl(self):
        raise NotImplementedError()

    @abstractmethod
    def __add__(self, other):
        raise NotImplementedError()

    def __iadd__(self, other):
        """
        __iadd__ is explicitly overriden to enforce that it is non-mutating.
        """
        return self + other


class IO(IOInterface):
    """
    Class for creating an interface bundle.

    @kwargs: ordered dict of {name: type}, ala decl.
    """
    # Note that because we require kwargs to be ordered, we have a strong
    # requirement here for >= python version 3.6. See
    # https://www.python.org/dev/peps/pep-0468/.
    def __init__(self, **kwargs):
        self._ports = {}
        self._decl = []
        for name, typ in kwargs.items():
            ref = LazyDefnRef(name=name)
            if isinstance(typ, MagmaProtocolMeta):
                port = typ.flip()
                port = port._from_magma_value_(port._to_magma_()(name=ref))
            else:
                port = typ.flip()(name=ref)
            self._ports[name] = port
            self._decl += [name, typ]
            setattr(self, name, port)
        self._bound = False

    @property
    def ports(self):
        return self._ports.copy()

    def bind(self, defn):
        if self._bound:
            raise Exception("Can not bind IO multiple times")
        for port in self._ports.values():
            port.name.set_defn(defn)
        self._bound = True

    def decl(self):
        return self._decl

    def __add__(self, other):
        """
        Attempts to combine this IO and @other. Returns a new IO object with the
        combined ports, unless:
          * @other is not of type IOInterface, in which case a TypeError is
            raised
          * this or @other has already been bound, in which case an Exception is
            raised
          * this and @other have common port names, in which case an Exception
            is raised
        """
        if not isinstance(other, IOInterface):
            raise TypeError(f"unsupported operand type(s) for +: 'IO' and "
                            f"'{type(other).__name__}'")
        if self._bound or other._bound:
            raise Exception("Adding bound IO not allowed")
        if self._ports.keys() & other._ports.keys():
            raise Exception("Adding IO with duplicate port names not allowed")
        decl = self._decl + other._decl
        return IO(**dict(zip(decl[::2], decl[1::2])))



from .passes import DefinitionPass
from ..digital import Digital
from ..generator import Generator
from ..circuit import Circuit, DeclareCoreirCircuit
from ..t import In, Out
from ..digital import Digital
from ..array import Array
from ..tuple import Tuple
from ..bits import Bits
from ..ref import NamedRef


def _simulate_wire(self, value_store, state_store):
    value_store.set_value(self.O, value_store.get_value(self.I))


class Wire(Generator):
    @staticmethod
    def generate(T):
        if issubclass(T, Digital):
            coreir_lib = "corebit"
            coreir_genargs = None
        else:
            coreir_lib = "coreir"
            width = T.flat_length()
            T = Bits[width]
            coreir_genargs = {"width": width}
        return DeclareCoreirCircuit(
            "Wire", 'I', In(T), 'O', Out(T),
            simulate=_simulate_wire,
            coreir_name="wire",
            coreir_lib=coreir_lib,
            coreir_genargs=coreir_genargs
        )


def _sanitize_name(name):
    return name.replace("[", "_").replace("]", "")


class InsertCoreIRWires(DefinitionPass):
    def __init__(self, main):
        super().__init__(main)
        self.seen = set()
        self.wire_map = {}

    def insert_wire(self, value, definition):
        if value.is_mixed():
            # mixed children
            for child in value:
                if not child.is_output():
                    self.insert_wire(child, definition)
            return
        if value in self.seen:
            # In the case of inouts, we may see more than once
            return
        self.seen.add(value)
        driver = value.value()

        while driver is not None and driver.name.anon():
            if isinstance(driver, (Array, Tuple)) and \
                    not driver.iswhole(driver.ts):
                for child in value:
                    self.insert_wire(child, definition)
                return
            driver = driver.value()
        if driver is None or driver.is_output() or driver.is_inout():
            return

        value.unwire(driver)
        T = type(driver)
        if driver not in self.wire_map:
            driver_name = driver.name.qualifiedname("_")
            value_name = value.name.qualifiedname("_")
            driver_name = _sanitize_name(driver_name)
            value_name = _sanitize_name(value_name)

            # Rename driver so it doesn't conflict with new wire instance name
            if isinstance(driver.name, NamedRef):
                driver.name.name = "_" + driver.name.name
            name = f"{driver_name}"
            with definition.open():
                wire_inst = Wire(T, name=name)
                self.wire_map[driver] = wire_inst
                if issubclass(T, Digital):
                    wire_inst.I @= driver
                else:
                    wire_inst.I @= driver.as_bits()
        else:
            wire_inst = self.wire_map[driver]

        if issubclass(T, Digital):
            value @= wire_inst.O
        else:
            value @= T.from_bits(wire_inst.O)

        self.insert_wire(driver, definition)

    def __call__(self, definition):
        # Copy instances because inserting wire will append to instances
        instances_copy = definition.instances.copy()
        for instance in instances_copy:
            for value in instance.interface.ports.values():
                if not value.is_output():
                    self.insert_wire(value, definition)
        for value in definition.interface.ports.values():
            if not value.is_output():
                self.insert_wire(value, definition)



from functools import wraps, partial
import warnings


def deprecated(func=None, *, msg=None):
    """
    Inspired by https://pybit.es/decorator-optional-argument.html.

    Marks a function as deprecated. @msg is an optional parameter to override
    the default warning.

    Examples:

        @deprecated
        def foo(...): ...

        @deprecated('Don't use bar!')
        def bar(...): ...
    """
    if func is None:
        return partial(deprecated, msg=msg)
    if msg is None:
        msg = f"{func.__name__} is deprecated"

    @wraps(func)
    def _wrapper(*args, **kwargs):
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    return _wrapper



from coreir import Context
from ..array import Array
from ..tuple import Tuple
from ..clock import AsyncReset, AsyncResetN
from ..config import config, EnvConfig
from ..logging import root_logger
from ..t import In, Out
from .. import singleton
from ..circuit import DeclareCircuit
from .coreir_transformer import DefnOrDeclTransformer
from ..passes import DefinitionPass
from .util import keydefaultdict
from ..wire import wire


config._register(
    coreir_backend_log_level=EnvConfig(
        "MAGMA_COREIR_BACKEND_LOG_LEVEL", "WARN"),
)

_logger = root_logger().getChild("coreir_backend")
_logger.setLevel(config.coreir_backend_log_level)


# Singleton context meant to be used with coreir/magma code
@singleton
class CoreIRContextSingleton:
    __instance = None

    def get_instance(self):
        return self.__instance

    def reset_instance(self):
        self.__instance = Context()

    def __init__(self):
        self.__instance = Context()


CoreIRContextSingleton()

_context_to_modules = {}


class CoreIRBackend:
    def __init__(self, context=None):
        singleton = CoreIRContextSingleton().get_instance()
        if context is None:
            context = singleton
        if context is not singleton:
            _logger.warning("Creating CoreIRBackend with non-singleton CoreIR "
                            "context.")
        self.modules = _context_to_modules.setdefault(context, {})
        self.context = context
        self.libs = keydefaultdict(self.context.get_lib)
        self.libs_used = set()
        self.constant_cache = {}
        self.sv_bind_files = {}

    def compile(self, defn_or_decl):
        _logger.debug(f"Compiling: {defn_or_decl.name}")
        transformer = DefnOrDeclTransformer(self, defn_or_decl)
        transformer.run()
        self.modules[defn_or_decl.name] = transformer.coreir_module
        return self.modules


class InsertWrapCasts(DefinitionPass):
    def sim(self, value_store, state_store):
        input_val = value_store.get_value(getattr(self, "in"))
        value_store.set_value(self.out, input_val)

    def define_wrap(self, wrap_type, in_type, out_type):
        name = f"coreir_wrap{wrap_type}".replace("(", "").replace(")", "")
        return DeclareCircuit(name,
                              "in",
                              In(in_type),
                              "out",
                              Out(out_type),
                              coreir_genargs={"type": wrap_type},
                              coreir_name="wrap",
                              coreir_lib="coreir",
                              simulate=self.sim)

    def wrap_if_arst(self, port, definition):
        if isinstance(port, (Array, Tuple)):
            for t in port:
                self.wrap_if_arst(t, definition)
        elif port.is_input():
            if isinstance(port, (AsyncReset, AsyncResetN)) or \
                    isinstance(port.trace(), (AsyncReset, AsyncResetN)):
                value = port.trace()
                if value is not None and not issubclass(
                        type(value), type(port).flip()):
                    port.unwire(value)
                    if isinstance(port, (AsyncReset, AsyncResetN)):
                        inst = self.define_wrap(
                            type(port).flip(), type(port), type(value))()
                    else:
                        inst = self.define_wrap(
                            type(value).flip(), type(port), type(value))()
                    definition.place(inst)
                    getattr(inst, "in") <= value
                    wire(inst.out, port)

    def __call__(self, definition):
        # copy, because wrapping might add instances
        instances = definition.instances[:]
        for instance in definition.instances:
            if type(instance).coreir_name == "wrap" or \
                    type(instance).coreir_name == "unwrap":
                continue
            for port in instance.interface.ports.values():
                self.wrap_if_arst(port, definition)
        for port in definition.interface.ports.values():
            self.wrap_if_arst(port, definition)


def compile(main, file_name=None, context=None):
    InsertWrapCasts(main).run()
    backend = CoreIRBackend(context)
    backend.compile(main)
    if file_name is not None:
        return backend.modules[main.coreir_name].save_to_file(file_name)
    else:
        return backend.modules[main.coreir_name]



import ast
import enum
import textwrap
import inspect
from functools import wraps
from collections import namedtuple
import os

import six
from .config import get_compile_dir, set_compile_dir
# TODO: Remove circular dependency required for `circuit.bind` logic
import magma as m
from . import cache_definition
from .clock import ClockTypes
from .common import deprecated
from .interface import *
from .wire import *
from .config import get_debug_mode
from .debug import get_callee_frame_info, debug_info
from .logging import root_logger
from .is_definition import isdefinition
from .ref import AnonRef, ArrayRef, TupleRef, DefnRef, InstRef
from .bit import VCC, GND
from .array import Array
from .placer import Placer, StagedPlacer
from .tuple import Tuple
from .digital import Digital
from magma.syntax.combinational import combinational
from magma.syntax.sequential import sequential
from magma.syntax.verilog import combinational_to_verilog, \
    sequential_to_verilog
from .verilog_utils import value_to_verilog_name
from .view import PortView

__all__ = ['AnonymousCircuitType']
__all__ += ['AnonymousCircuit']

__all__ += ['CircuitType']
__all__ += ['Circuit']
__all__ += ['DeclareCircuit']
__all__ += ['DefineCircuit', 'EndDefine', 'EndCircuit']

__all__ += ['CopyInstance']
__all__ += ['circuit_type_method']
__all__ += ['circuit_generator']

circuit_type_method = namedtuple('circuit_type_method', ['name', 'definition'])

_logger = root_logger()


class _SyntaxStyle(enum.Enum):
    NONE = enum.auto()
    OLD = enum.auto()
    NEW = enum.auto()


# Maintain a stack of nested definitions.
class _PlacerBlock:
    __stack = []

    def __init__(self, placer):
        self.__placer = placer

    def __enter__(self):
        _PlacerBlock.push(self.__placer)

    def __exit__(self, typ, value, traceback):
        _PlacerBlock.pop()

    @classmethod
    def push(cls, placer):
        cls.__stack.append(placer)

    @classmethod
    def pop(cls):
        if not cls.__stack:
            return None
        return cls.__stack.pop()

    @classmethod
    def peek(cls):
        if not cls.__stack:
            return None
        return cls.__stack[-1]


def _setattrs(obj, dct):
    for k, v in dct.items():
        if isinstance(k, str):
            setattr(obj, k, v)


def _has_definition(cls, port=None):
    if cls.instances:
        return True
    if port is None:
        interface = getattr(cls, "interface", None)
        if not interface:
            return None
        return any(_has_definition(cls, p) for p in interface.ports.values())
    if isinstance(port, Tuple) or isinstance(port, Array):
        return any(_has_definition(cls, elem) for elem in port)
    if not port.is_output():
        return port.value() is not None
    return False


def _get_interface_decl(cls):
    if hasattr(cls, "IO") and hasattr(cls, "io"):
        _logger.warning("'IO' and 'io' should not both be specified, ignoring "
                        "'io'", debug_info=cls.debug_info)
    # TODO(setaluri): Simplify this logic and remove InterfaceKind check
    # (this is a artifact of the circuit_generator decorator).
    if hasattr(cls, 'IO'):
        cls._syntax_style_ = _SyntaxStyle.OLD
        _logger.warning("'IO = [...]' syntax is deprecated, use "
                        "'io = IO(...)' syntax instead",
                        debug_info=cls.debug_info)
        if isinstance(cls.IO, InterfaceKind):
            return None
        return DeclareInterface(*cls.IO)
    if hasattr(cls, "io"):
        cls._syntax_style_ = _SyntaxStyle.NEW
        return DeclareLazyInterface(cls.io)
    return None


def _add_intermediate_value(value, values):
    """
    Add an intermediate value `value` to `values`, handling members of
    recursive types.  Used by `_get_intermediate_values` as part of
    `CircuitKind.__repr__`.
    """
    # If we encounter a member of an array or tuple, add the entire parent
    # value (only once)
    if isinstance(value.name, ArrayRef):
        _add_intermediate_value(value.name.array, values)
    elif isinstance(value.name, TupleRef):
        _add_intermediate_value(value.name.tuple, values)
    elif not isinstance(value.name, (DefnRef, InstRef, AnonRef)):
        if value is VCC or value is GND:
            # Skip VCC and GND because they are special
            return
        if not any(value is x for x in values):
            values.append(value)


def _get_intermediate_values(value, values):
    """
    Retrieve the intermediate values connected to `value` and put them into
    `values`

    Used in the implementation of `CircuitKind.__repr__` to emit the temporary
    values in a circuit.
    """
    if value.is_output():
        return
    if value.is_mixed():
        # Mixed
        for v in value:
            _get_intermediate_values(v, values)
        return
    driver = value.value()
    if driver is None:
        return
    if isinstance(value, (Array, Tuple)) and driver.name.anon():
        for elem in value:
            _get_intermediate_values(elem, values)
    else:
        while driver is not None:
            _add_intermediate_value(driver, values)
            if not driver.is_output():
                value = driver
                driver = driver.value()
            else:
                driver = None


class CircuitKind(type):
    def __prepare__(name, bases, **kwargs):
        _PlacerBlock.push(StagedPlacer(name))
        return type.__prepare__(name, bases, **kwargs)

    """Metaclass for creating circuits."""
    def __new__(metacls, name, bases, dct):
        # Override class name if supplied (and save class name).
        cls_name = dct.get("_cls_name_", name)
        name = dct.setdefault('name', name)

        dct.setdefault('renamed_ports', {})
        dct.setdefault('primitive', False)
        dct.setdefault('coreir_lib', 'global')
        dct["inline_verilog_strs"] = []
        dct["bind_modules"] = {}

        # If in debug_mode is active and debug_info is not supplied, attach
        # callee stack info.
        dct.setdefault("debug_info", None)
        if get_debug_mode() and not dct["debug_info"]:
            callee_frame = inspect.getframeinfo(
                inspect.currentframe().f_back.f_back)
            module = inspect.getmodule(inspect.stack()[2][0])
            dct["debug_info"] = debug_info(callee_frame.filename,
                                           callee_frame.lineno, module)

        cls = type.__new__(metacls, name, bases, dct)

        for method in dct.get('circuit_type_methods', []):
            setattr(cls, method.name, method.definition)

        # Create interface for this circuit class.
        cls._syntax_style_ = _SyntaxStyle.NONE
        IO = _get_interface_decl(cls)
        if IO is not None:
            cls.IO = IO
            cls.interface = cls.IO(defn=cls,
                                   renamed_ports=dct["renamed_ports"])
            _setattrs(cls, cls.interface.ports)
        placer = _PlacerBlock.pop()
        if placer:
            assert placer.name == cls_name
            cls._placer = placer.finalize(cls)
        else:
            cls._placer = Placer(cls)

        return cls

    def __call__(cls, *largs, **kwargs):
        if get_debug_mode():
            debug_info = get_callee_frame_info()
            kwargs["debug_info"] = debug_info
        self = super(CircuitKind, cls).__call__(*largs, **kwargs)
        if hasattr(cls, 'IO'):
            interface = cls.IO(inst=self, renamed_ports=cls.renamed_ports)
            self.setinterface(interface)
        return self

    def __str__(cls):
        interface = ""
        if hasattr(cls, "interface"):
            interface = ", ".join(f"{name}: {_type}"
                                  for name, _type in cls.interface.items())
            interface = f"({interface})"
        return f"{cls.__name__}{interface}"

    def __repr__(cls):
        if not hasattr(cls, 'IO'):
            return super().__repr__()

        name = cls.__name__
        args = str(cls.IO)

        if not hasattr(cls, "instances"):
            return f"{name} = DeclareCircuit(\"{name}\", {args})"
        s = f"{name} = DefineCircuit(\"{name}\", {args})\n"
        sorted_instances = sorted(cls.instances, key=lambda x: x.name)

        values = []
        for instance in sorted_instances:
            values.extend(list(instance.interface.ports.values()))
        values.extend(cls.interface.ports.values())
        intermediate_values = []
        for value in values:
            _get_intermediate_values(value, intermediate_values)
        if intermediate_values:
            s += "\n".join(
                f"{value.name} = {repr(value)}" for value in intermediate_values
            ) + "\n"

        # Emit instances.
        for instance in sorted_instances:
            s += repr(instance) + '\n'
        # Emit wires from instances.
        for instance in sorted_instances:
            s += repr(instance.interface)
        s += repr(cls.interface)
        s += "EndCircuit()"
        return s

    def rename(cls, new_name):
        if cls.verilogFile:
            raise Exception("Can not rename a verilog wrapped file")

        old_name = cls.name
        cls.name = new_name
        cls.coreir_name = new_name
        cls.verilog_name = new_name
        cls.__name__ = new_name

        # NOTE(rsetaluri): This is a very hacky way to try to rename wrapped
        # verilog. We simply replace the first instance of "module <old_name>"
        # with "module <new_name>". This ignores the possibility of "module
        # <new_name>" existing anywhere else, most likely in comments etc. The
        # more robust way to do this would to modify the AST directly and
        # generate the new verilog code.
        #
        # Instead, at the top of this function, we raise an exception for
        # verilog wrapped files. That is, for now it is considered an error to
        # try to rename a verilog wrapped circuit.
        if cls.verilogFile:
            find_str = f"module {old_name}"
            replace_str = f"module {new_name}"
            assert cls.verilogFile.find(find_str) != -1
            cls.verilogFile = cls.verilogFile.replace(find_str, replace_str, 1)

    def find(cls, defn):
        name = cls.__name__
        if not isdefinition(cls):
            return defn
        for i in cls.instances:
            type(i).find(defn)
        if name not in defn:
            defn[name] = cls
        return defn

    def inline_verilog(cls, inline_str, **kwargs):
        format_args = {}
        for key, arg in kwargs.items():
            if isinstance(arg, m.Type):
                # TODO: For now we assumed values aren't used elswhere, this
                # forces it to be connected to an instance, so temporaries will
                # appear in the code
                if arg.is_inout() and not arg.driven():
                    raise NotImplementedError()
                if arg.is_input() and not arg.driven():
                    # TODO: Could be driven after, but that will just override
                    # this wiring so it's okay for now
                    arg.undriven()
                elif arg.is_output() and not arg.wired():
                    arg.unused()
                elif not (arg.is_input() or arg.is_output() or arg.is_inout()):
                    arg.unused()
                arg = value_to_verilog_name(arg)
            elif isinstance(arg, PortView):
                arg = value_to_verilog_name(arg)
            format_args[key] = arg
        cls.inline_verilog_strs.append(
            inline_str.format(**format_args))


@six.add_metaclass(CircuitKind)
class AnonymousCircuitType(object):
    """Abstract base class for circuits"""
    def __init__(self, *largs, **kwargs):
        self.kwargs = dict(**kwargs)

        if "debug_info" in self.kwargs:
            # Not an instance parameter, internal debug argument.
            del self.kwargs["debug_info"]

        if hasattr(self, 'default_kwargs'):
            for key, value in self.default_kwargs.items():
                self.kwargs.setdefault(key, value)

        self.name = kwargs['name'] if 'name' in kwargs else ""
        self.loc = kwargs['loc'] if 'loc' in kwargs else None
        if self.loc and len(self.loc) == 2:
            self.loc = (self.loc[0], self.loc[1], 0)
        self.interface = None
        self.defn = None
        self.used = False
        self.is_instance = True
        self.debug_info = kwargs.get("debug_info", None)

    def set_debug_info(self, debug_info):
        self.debug_info = debug_info

    def __str__(self):
        if self.name:
            return f"{self.name}<{type(self)}>"
        name = f"AnonymousCircuitInst{id(self)}"
        interface = ", ".join(f"{name}: {type(value)}"
                              for name, value in self.interface.ports.items())
        return f"{name}<{interface}>"

    def __repr__(self):
        args = []
        for k, v in self.interface.ports.items():
            args.append(f"\"{k}\"")
            args.append(repr(v))
        args = ", ".join(args)
        typ = type(self).__name__
        if self.name:
            return f"{self.name} = {typ}({args})"
        return f"{typ}({args})"

    def __getitem__(self, key):
        return self.interface[key]

    def wireoutputs(self, outputs, debug_info):
        """Wire a list of outputs to the circuit's inputs"""
        inputs = self.interface.inputs()
        ni = len(inputs)
        no = len(outputs)
        if ni != no:
            msg = (
                f"Number of inputs is not equal to the number of outputs, "
                f"expected {ni} inputs, got {no}. Only {min(ni,no)} will be "
                f"wired.")
            _logger.warning(msg, debug_info=debug_info)
        for i in range(min(ni, no)):
            wire(outputs[i], inputs[i], debug_info)

    def wire(self, output, debug_info):
        """Wire a single output to the circuit's inputs"""
        if hasattr(output, 'interface'):
            # Wire the circuit's outputs to this circuit's inputs.
            self.wireoutputs(output.interface.outputs(), debug_info)
            return
        if isinstance(output, m.MagmaProtocol):
            output = output._get_magma_value_()
        # Wire the output to this circuit's input (should only have 1 input).
        inputs = []
        for inp in self.interface.inputs():
            if isinstance(inp, m.MagmaProtocol):
                inp = inp._get_magma_value_()
            inputs.append(inp)
        ni = len(inputs)
        if ni == 0:
            msg = ("Wiring an output to a circuit with no input arguments, "
                   "skipping")
            _logger.warning(msg, debug_info=debug_info)
            return
        if ni != 1:
            msg = (f"Wiring an output to a circuit with more than one input "
                   f"argument, using the first input {inputs[0].debug_name}")
            _logger.warning(msg, debug_info=debug_info)
        inputs[0].wire(output, debug_info)

    @property
    def debug_name(self):
        defn_str = ""
        if hasattr(self, 'defn') and self.defn is not None:
            defn_str = str(self.defn.name)
        return f"{defn_str}.{self.name}"

    def __call__(input, *outputs, **kw):
        debug_info = None
        if get_debug_mode():
            debug_info = get_callee_frame_info()

        no = len(outputs)
        if len(outputs) == 1:
            input.wire(outputs[0], debug_info)
        elif len(outputs) >= 1:  # In case there are only kw
            input.wireoutputs(outputs, debug_info)

        # Wire up extra arguments, name to name.
        # TODO(rsetaluri): This code should be changed to use clock types.
        for key, value in kw.items():
            if key == 'enable':
                key = 'CE'
            elif key == 'reset':
                key = 'RESET'
            elif key == 'set':
                key = 'SET'  # NYI
            elif key == 'ce':
                key = 'CE'  # deprecated
            if hasattr(input, key):
                i = getattr(input, key)
                wire(value, getattr(input, key), debug_info)
            else:
                msg = f"Instance {input.debug_name} does not have input {key}"
                _logger.warning(msg, debug_info=debug_info)

        o = input.interface.outputs()
        return o[0] if len(o) == 1 else tuple(o)

    def setinterface(self, interface):
        _setattrs(self, interface.ports)
        self.interface = interface
        return self

    def on(self):
        self.used = True
        return self

    def off(self):
        self.used = False
        return self

    def rename(self, name):
        self.name = name
        return self

    def isclocked(self):
        return self.interface.isclocked()

    def clockargs(self):
        return self.interface.clockargs()

    def inputargs(self):
        return self.interface.inputargs()

    def outputargs(self):
        return self.interface.outputargs()

    @classmethod
    def open(cls):
        return _PlacerBlock(cls._placer)


def AnonymousCircuit(*decl):
    """
    AnonymousCircuits are like macros - the circuit instances are not placed
    """
    if len(decl) == 1:
        decl = decl[0]
    return AnonymousCircuitType().setinterface(AnonymousInterface(decl))


class CircuitType(AnonymousCircuitType):
    """Placed circuit - instances placed in a definition"""
    def __init__(self, *largs, **kwargs):
        super(CircuitType, self).__init__(*largs, **kwargs)
        # Circuit instances are placed if within a definition.
        placer = _PlacerBlock.peek()
        if placer:
            placer.place(self)

    def __repr__(self):
        args = []
        for k, v in self.kwargs.items():
            if isinstance(v, tuple):
                # {   # Format identifier
                # 0:  # first parameter
                # #   # use "0x" prefix
                # 0   # fill with zeroes
                # {1} # to n characters (including 0x), per the second parameter
                # x   # hexadecimal number, using lowercase letters for a-f
                # }   # End of format identifier
                if len(v) == 2:
                    v = "{0:#0{1}x}".format(v[0], v[1] // 4)
            else:
                v = f"\"{v}\""
            args.append(f"{k}={v}")
        args = ", ".join(args)
        typ = type(self).__name__
        if self.name:
            return f"{self.name} = {typ}({args})"
        return f"{typ}({args})"


@deprecated(
    msg="DeclareCircuit factory method is deprecated, subclass Circuit instead")
def DeclareCircuit(name, *decl, **args):
    """DeclareCircuit Factory"""
    debug_info = None
    if get_debug_mode():
        debug_info = get_callee_frame_info()
    metacls = CircuitKind
    bases = (CircuitType,)
    dct = metacls.__prepare__(name, bases)
    dct.update(dict(IO=decl,
                    debug_info=debug_info,
                    is_definition=False,
                    primitive=args.get('primitive', True),
                    stateful=args.get('stateful', False),
                    simulate=args.get('simulate'),
                    firrtl_op=args.get('firrtl_op'),
                    circuit_type_methods=args.get('circuit_type_methods', []),
                    coreir_lib=args.get('coreir_lib', "global"),
                    coreir_name=args.get('coreir_name', name),
                    coreir_genargs=args.get('coreir_genargs', None),
                    coreir_configargs=args.get('coreir_configargs', {}),
                    verilog_name=args.get('verilog_name', name),
                    default_kwargs=args.get('default_kwargs', {}),
                    renamed_ports=args.get('renamed_ports', {})))
    return metacls(name, bases, dct)


class DefineCircuitKind(CircuitKind):
    def __new__(metacls, name, bases, dct):
        dct["_cls_name_"] = name  # save original name for debugging purposes
        if 'name' not in dct:
            dct['name'] = name
            # Check if we are a subclass of something other than Circuit; if so,
            # inherit the name of the first parent.
            for base in bases:
                if base is Circuit:
                    continue
                if not issubclass(base, Circuit):
                    raise Exception(f"Must subclass from Circuit or a "
                                    f"subclass of Circuit ({base})")
                dct['name'] = base.name
                break
        name = dct['name']
        dct["renamed_ports"] = dct.get("renamed_ports", {})

        self = CircuitKind.__new__(metacls, name, bases, dct)

        self.verilog = None
        self.verilogFile = None
        self.verilogLib = None

        self.verilog_name = dct.get('verilog_name', name)
        self.coreir_name = dct.get('coreir_name', name)
        self.coreir_genargs = dct.get('coreir_genargs', None)
        self.coreir_configargs = dct.get('coreir_configargs', {})
        self.default_kwargs = dct.get('default_kwargs', {})
        self.firrtl = None

        has_definition = _has_definition(self)
        self._is_definition = dct.get('is_definition', has_definition)
        self.is_instance = False

        run_unconnected_check = False
        if hasattr(self, "definition"):
            if self._syntax_style_ is _SyntaxStyle.OLD:
                _logger.warning("'definition' class method syntax is "
                                "deprecated, use inline definition syntax "
                                "instead", debug_info=self.debug_info)
                with _PlacerBlock(self._placer):
                    self.definition()
                self._is_definition = True
                run_unconnected_check = True
            elif self._syntax_style_ is _SyntaxStyle.NEW:
                _logger.warning("Supplying method 'definition' with new inline "
                                "definition syntax is not supported, ignoring "
                                "'definition'")
        run_unconnected_check = run_unconnected_check or (
            has_definition and self._syntax_style_ is _SyntaxStyle.NEW)
        run_unconnected_check = run_unconnected_check and not \
            dct.get("_ignore_undriven_", False)
        if run_unconnected_check:
            self.check_unconnected()

        return self

    def check_unconnected(self):
        for port in self.interface.inputs():
            if not port.driven():
                msg = f"Output port {self.name}.{port.name} not driven"
                _logger.error(msg, debug_info=self.debug_info)

        for inst in self.instances:
            for port in inst.interface.inputs():
                if not port.driven():
                    msg = f"Input port {inst.name}.{port.name} not driven"
                    _logger.error(msg, debug_info=inst.debug_info)

    @property
    def is_definition(self):
        return self._is_definition or self.verilog or self.verilogFile

    @property
    def instances(self):
        return self._placer.instances()

    def place(cls, inst):
        """Place a circuit instance in this definition"""
        cls._placer.place(inst)

    def gen_bind_port(cls, mon_arg, bind_arg):
        if isinstance(mon_arg, Tuple) or isinstance(mon_arg, Array) and \
                not issubclass(mon_arg.T, m.Digital):
            result = []
            for child1, child2 in zip(mon_arg, bind_arg):
                result += cls.gen_bind_port(child1, child2)
            return result
        port = value_to_verilog_name(mon_arg)
        arg = value_to_verilog_name(bind_arg)
        return [(f".{port}({arg})")]

    def bind(cls, monitor, *args):
        bind_str = monitor.verilogFile

        ports = []
        for mon_arg, cls_arg in zip(monitor.interface.ports.values(),
                                    cls.interface.ports.values()):
            if str(mon_arg.name) != str(cls_arg.name):
                error_str = f"""
Bind monitor interface does not match circuit interface
    Monitor Ports: {list(monitor.interface.ports)}
    Circuit Ports: {list(cls.interface.ports)}
"""
                raise TypeError(error_str)
            ports += cls.gen_bind_port(mon_arg, cls_arg)
        extra_mon_args = list(
            monitor.interface.ports.values()
        )[len(cls.interface):]
        for mon_arg, bind_arg in zip(extra_mon_args, args):
            ports += cls.gen_bind_port(mon_arg, bind_arg)
        ports_str = ", ".join(ports)
        bind_str = f"bind {cls.name} {monitor.name} {monitor.name}_inst ({ports_str});"  # noqa
        if not os.path.isdir(".magma"):
            os.mkdir(".magma")
        curr_compile_dir = get_compile_dir()
        set_compile_dir("normal")
        # Circular dependency, need coreir backend to compile, backend imports
        # circuit (for wrap casts logic, we might be able to factor that out)
        m.compile(f".magma/{monitor.name}", monitor, inline=True)
        set_compile_dir(curr_compile_dir)
        with open(f".magma/{monitor.name}.v", "r") as f:
            content = "\n".join((f.read(), bind_str))
        cls.bind_modules[monitor.name] = content


@six.add_metaclass(DefineCircuitKind)
class Circuit(CircuitType):
    pass


@deprecated(
    msg="DefineCircuit factory method is deprecated, subclass Circuit instead")
def DefineCircuit(name, *decl, **args):
    """DefineCircuit Factory"""
    debug_info = None
    if get_debug_mode():
        debug_info = get_callee_frame_info()
    metacls = DefineCircuitKind
    bases = (Circuit,)
    dct = metacls.__prepare__(name, bases)
    dct.update(dict(IO=decl,
                    is_definition=True,
                    primitive=args.get('primitive', False),
                    stateful=args.get('stateful', False),
                    simulate=args.get('simulate'),
                    debug_info=debug_info,
                    verilog_name=args.get('verilog_name', name),
                    coreir_name=args.get('coreir_name', name),
                    coreir_lib=args.get('coreir_lib', "global"),
                    coreir_genargs=args.get('coreir_genargs', None),
                    coreir_configargs=args.get('coreir_configargs', None),
                    default_kwargs=args.get('default_kwargs', {}),
                    renamed_ports=args.get('renamed_ports', {}),
                    kratos=args.get("kratos", None)))
    defn = metacls(name, bases, dct)
    _PlacerBlock.push(defn._placer)
    return defn


def EndDefine():
    placer = _PlacerBlock.pop()
    if not placer:
        raise Exception("EndDefine called without DefineCircuit")
    placer._defn.check_unconnected()
    debug_info = get_callee_frame_info()
    placer._defn.end_circuit_filename = debug_info[0]
    placer._defn.end_circuit_lineno = debug_info[1]


EndCircuit = EndDefine


def CopyInstance(instance):
    circuit = type(instance)
    new_instance = circuit()
    new_instance.kwargs = instance.kwargs
    return new_instance


GeneratorArguments = namedtuple('GeneratorArguments', ['args', 'kwargs'])


def circuit_generator(func):
    @cache_definition
    @wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        # Store arguments to generate the circuit.
        result._generator_arguments = GeneratorArguments(args, kwargs)
        return result
    return wrapped


coreir_port_mapping = {
    "I": "in",
    "I0": "in0",
    "I1": "in1",
    "O": "out",
    "S": "sel",
    "CLK": "clk",
}


def DeclareCoreirCircuit(*args, **kwargs):
    return DeclareCircuit(*args, **kwargs,
                          renamed_ports=coreir_port_mapping)



import magma as m


primitive_to_python_operator_name_map = {
    "and": "and_",
    "or": "or_",
    "xor": "xor",
    "shl": "lshift",
    "lshr": "rshift",
    "ashr": "rshift",
    "urem": "mod",
    "srem": "mod",
    "udiv": "floordiv",
    "sdiv": "floordiv",
    "ule": "le",
    "ult": "lt",
    "uge": "ge",
    "ugt": "gt",
    "sle": "le",
    "slt": "lt",
    "sge": "ge",
    "sgt": "gt",
    "not": "invert"
}


def BitOrBits(width):
    if width is None:
        return m.Bit
    if not isinstance(width, int):
        raise ValueError(f"Expected width to be None or int, got {width}")
    return m.Bits[width]


def pretty_str(t):
    if issubclass(t, m.Tuple):
        args = []
        for i in range(len(t)):
            key_str = str(list(t.keys())[i])
            val_str = pretty_str(list(t.types())[i])
            indent = " " * 4
            val_str = f"\n{indent}".join(val_str.splitlines())
            args.append(f"{key_str} = {val_str}")
        # Pretty print by using newlines + indent
        joiner = ",\n    "
        result = joiner.join(args)
        # Insert first newline + indent and last newline
        result = "\n    " + result + "\n"
        s = f"Tuple({result})"
    elif issubclass(t, m.Bits):
        s = str(t)
    elif issubclass(t, m.Array):
        s = f"Array[{t.N}, {pretty_str(t.T)}]"
    else:
        s = str(t)
    return s


def pretty_print_type(t):
    print(pretty_str(t))



from .t import Direction, In
from .digital import DigitalMeta, Digital
from .wire import wire


class Clock(Digital, metaclass=DigitalMeta):
    pass


ClockIn = Clock[Direction.In]
ClockOut = Clock[Direction.Out]


# synchronous reset, active high (i.e. reset when signal is 1)
class Reset(Digital, metaclass=DigitalMeta):
    pass


ResetIn = Reset[Direction.In]
ResetOut = Reset[Direction.Out]


# synchronous reset, active low (i.e. reset when signal is 0)
class ResetN(Digital, metaclass=DigitalMeta):
    pass


ResetNIn = ResetN[Direction.In]
ResetNOut = ResetN[Direction.Out]


# asynchronous reset, active high (i.e. reset when signal is 1)
class AsyncReset(Digital, metaclass=DigitalMeta):
    pass


AsyncResetIn = AsyncReset[Direction.In]
AsyncResetOut = AsyncReset[Direction.Out]


# asynchronous reset, active low (i.e. reset when signal is 0)
class AsyncResetN(Digital, metaclass=DigitalMeta):
    pass


AsyncResetNIn = AsyncResetN[Direction.In]
AsyncResetNOut = AsyncResetN[Direction.Out]


# Preset
# Clear
class Enable(Digital, metaclass=DigitalMeta):
    pass


EnableIn = Enable[Direction.In]
EnableOut = Enable[Direction.Out]

ClockTypes = (Clock, Reset, AsyncReset, Enable)


def ClockInterface(has_enable=False, has_reset=False, has_ce=False,
                   has_async_reset=False, has_async_resetn=False):
    args = ['CLK', In(Clock)]
    has_enable |= has_ce
    if has_enable:
        args += ['CE', In(Enable)]
    if has_reset:
        args += ['RESET', In(Reset)]
    if has_async_reset:
        args += ['ASYNCRESET', In(AsyncReset)]
    if has_async_resetn:
        args += ['ASYNCRESETN', In(AsyncResetN)]
    return args


def wireclocktype(defn, inst, clocktype):
    defnclk = []
    for port in defn.interface.ports.values():
        if isinstance(port, clocktype) and port.is_output():
            defnclk += [port]
    if defnclk:
        defnclk = defnclk[0]  # wire first clock
        for port in inst.interface.inputs(include_clocks=True):
            if isinstance(port, clocktype) and not port.driven():
                wire(defnclk, port)


def wiredefaultclock(defn, inst):
    wireclocktype(defn, inst, Clock)


def wireclock(define, circuit):
    wireclocktype(define, circuit, Reset)
    wireclocktype(define, circuit, AsyncReset)
    wireclocktype(define, circuit, AsyncResetN)
    wireclocktype(define, circuit, Enable)



from .t import In
from .generator import Generator


def MonitorIO(circuit):
    IO = []
    for name, port in circuit.interface.ports.items():
        IO += [name, In(type(port))]
    return IO


class MonitorGenerator:
    pass



import magma as m
from .compatibility import IntegerTypes
from .debug import debug_wire
from .logging import root_logger
from .t import MagmaProtocol


_logger = root_logger()


@debug_wire
def wire(o, i, debug_info=None):
    if isinstance(o, MagmaProtocol):
        o = o._get_magma_value_()
    if isinstance(i, MagmaProtocol):
        i = i._get_magma_value_()

    # Wire(o, Circuit).
    if hasattr(i, 'interface'):
        i.wire(o, debug_info)
        return

    # Replace output Circuit with its output (should only be 1 output).
    if hasattr(o, 'interface'):
        # If wiring a Circuit to a Port then circuit should have 1 output.
        o_orig = o
        o = o.interface.outputs()
        if len(o) != 1:
            _logger.error(f"Can only wire circuits with one output. Argument "
                          f"0 to wire `{o_orig.debug_name}` has outputs {o}",
                          debug_info=debug_info)
            return
        o = o[0]

    # If o is an input.
    if not isinstance(o, IntegerTypes) and o.is_input():
        # If i is not an input.
        if isinstance(i, IntegerTypes) or not i.is_input():
            # Flip i and o.
            i, o = o, i

    # Wire(o, Type).
    i.wire(o, debug_info)


class Wire:
    """
    Wire implement wiring.

    Each wire is represented by a Bit().
    """
    def __init__(self, bit):
        self.bit = bit
        self.driving = []
        self.driver = None

    def __repr__(self):
        return repr(self.bit)

    def __str__(self):
        return str(self.bit)

    def anon(self):
        return self.bit.anon()

    def unwire(self, other):
        other.driving.remove(self)
        self.driver = None

    def connect(self, other, debug_info):
        """
        Connect two wires, self should be an input and other should be an
        output, or both should be inouts
        """
        if self.driver is not None and self.bit.is_input():
            _logger.warning(
                "Wiring multiple outputs to same wire, using last connection."
                f" Input: {self.bit.debug_name}, "
                f" Old Output: {self.driver.bit.debug_name}, "
                f" New Output: {other.bit.debug_name}",
                debug_info=debug_info
            )
        if self.bit.is_output():
            _logger.error(f"Using `{self.bit.debug_name}` (an output) as an "
                          f"input", debug_info=debug_info)
            return
        if other.bit.is_input():
            _logger.error(f"Using `{other.bit.debug_name}` (an input) as an "
                          f"output", debug_info=debug_info)
            return
        if self.bit.is_inout() and not other.bit.is_inout():
            _logger.error(f"Using `{other.bit.debug_name}` (not inout) as an "
                          f"inout", debug_info=debug_info)
            return
        if not self.bit.is_inout() and other.bit.is_inout():
            _logger.error(f"Using `{self.bit.debug_name}` (not inout) as an "
                          f"inout", debug_info=debug_info)
            return

        self.driver = other
        other.driving.append(self)

    def trace(self, skip_self=True):
        """
        If a value is an input or an intermediate (undirected), trace it until
        there is an input or inout (this is the source)

        Upon the first invocation (from a user), we skip the current bit (so
        we don't trace to ourselves)
        """
        if self.driver is not None:
            return self.driver.trace(skip_self=False)
        if not skip_self and (self.bit.is_output() or self.bit.is_inout()):
            return self.bit
        return None

    def value(self):
        """
        Return the driver of this wire
        """
        if self.bit.is_output():
            raise TypeError("Can only get value of non outputs")
        if self.driver is None:
            return None
        return self.driver.bit

    def driven(self):
        return self.trace() is not None

    def wired(self):
        return self.driver or self.driving



import functools
import warnings
import enum
from abc import abstractmethod
from .common import deprecated
from .ref import AnonRef, NamedRef, DefnRef, InstRef
from .compatibility import IntegerTypes, StringTypes


class Direction(enum.Enum):
    In = 0
    Out = 1
    InOut = 2
    Undirected = 3


class Type(object):
    def __init__(self, name=None):
        if name is None:
            name = AnonRef()
        elif isinstance(name, str):
            name = NamedRef(name=name)
        self.name = name

    __hash__ = object.__hash__

    def __repr__(self):
        if self.name.anon():
            return f"{type(self)}()"
        if isinstance(self.name, NamedRef) and \
                not isinstance(self.name, (InstRef, DefnRef)):
            return f"{type(self)}(name=\"{repr(self.name)}\")"
        return repr(self.name)

    def __str__(self):
        return str(self.name)

    # an instance has an anon name
    def anon(self):
        return self.name.anon()

    # abstract method - must be implemented by subclasses
    @classmethod
    def is_oriented(cls, direction):
        raise NotImplementedError()

    @classmethod
    def is_input(cls):
        return cls.is_oriented(Direction.In)

    @classmethod
    def is_output(cls):
        return cls.is_oriented(Direction.Out)

    @classmethod
    def is_inout(cls):
        return cls.is_oriented(Direction.InOut)

    @classmethod
    @deprecated
    def isoriented(cls, direction):
        return cls.is_oriented(direction)

    @classmethod
    @deprecated
    def isinput(cls):
        return cls.is_input()

    @classmethod
    @deprecated
    def isoutput(cls):
        return cls.is_output()

    @classmethod
    @deprecated
    def isinout(cls):
        return cls.is_inout()

    @property
    def debug_name(self):
        defn_str = ""
        inst_str = ""
        if isinstance(self.name, DefnRef):
            defn_str = str(self.name.defn.name) + "."
        elif isinstance(self.name, InstRef):
            inst_str = str(self.name.inst.name) + "."
            defn_str = str(self.name.inst.defn.name) + "."
        return f"{defn_str}{inst_str}{str(self)}"

    def __le__(self, other):
        if not self.is_output():
            self.wire(other)
        else:
            raise TypeError(f"Cannot use <= to assign to output: {self.debug_name} (trying to assign {other.debug_name})")

    def __imatmul__(self, other):
        if not self.is_output():
            self.wire(other)
        else:
            raise TypeError(f"Cannot use @= to assign to output: {self} (trying to assign {other})")
        return self

    @abstractmethod
    def unused(self):
        # Mark value is unused by calling unused on the underlying magma
        # elements
        # For example, m.Bit is wired up to a coreir term primitive
        # A general m.Array and m.Tuple will recursively call `unused` on its
        # members
        raise NotImplementedError()

    @abstractmethod
    def undriven(self):
        # Mark value is undriven by calling undriven on the underlying magma
        # elements
        # For example, m.Bit is wired up to a coreir undriven primitive
        # A general m.Array and m.Tuple will recursively call `undriven` on its
        # members
        raise NotImplementedError()

    @abstractmethod
    def as_bits(self):
        """
        Convert value to "flat" representation as an instance of m.Bits

        For example, flatten a multi-dimensional array `m.Array[4,
        m.Bits[5]]` into `m.Bits[20]`.

        This is the inverse of `from_bits`
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_bits(self):
        """
        Create value from a "flat" representation as an instance of m.Bits.

        For example, pack an `m.Bits[20]` into an `m.Array[4, m.Bits[5]]`.

        This is the inverse of `as_bits`
        """
        raise NotImplementedError()



class Kind(type):
    # subclasses only need to implement one of these methods
    def __eq__(cls, rhs):
        return cls is rhs

    __hash__ = type.__hash__

    def __repr__(cls):
        return cls.__name__

    def __str__(cls):
        return cls.__name__

    @abstractmethod
    def qualify(cls, direction):
        raise NotImplementedError()

    def flip(cls):
        if cls.direction == Direction.In:
            return cls[Direction.Out]
        elif cls.direction == Direction.Out:
            return cls[Direction.In]
        else:
            # Flip of inout is inout
            # Flip of undirected is undirected
            return cls


def In(T):
    return T.qualify(Direction.In)


def Out(T):
    return T.qualify(Direction.Out)


def InOut(T):
    return T.qualify(Direction.InOut)


def Flip(T):
    return T.flip()


class MagmaProtocolMeta(type):
    @abstractmethod
    def _to_magma_(cls):
        # Need way to retrieve underlying magma type
        raise NotImplementedError()

    @abstractmethod
    def _qualify_magma_(cls, direction: Direction):
        # Need way to qualify underlying type (e.g. give me a Foo with the
        # underlying type qualified to be an input)
        raise NotImplementedError()

    @abstractmethod
    def _flip_magma_(cls):
        # Need way to flip underlying type (e.g. give me a Foo with the
        # underlying type flipped)
        raise NotImplementedError()

    def qualify(cls, direction: Direction):
        return cls._qualify_magma_(direction)

    @abstractmethod
    def _from_magma_value_(cls, val: Type):
        # Need a way to create an instance from a value
        raise NotImplementedError()

    def flip(cls):
        return cls._flip_magma_()


class MagmaProtocol(metaclass=MagmaProtocolMeta):
    @abstractmethod
    def _get_magma_value_(self):
        # Need way to access underlying magma value
        raise NotImplementedError()

    @classmethod
    def is_input(cls):
        return cls._to_magma_().is_input()

    @classmethod
    def is_mixed(cls):
        return cls._to_magma_().is_mixed()

    @classmethod
    def is_output(cls):
        return cls._to_magma_().is_output()

    def value(cls):
        return cls._get_magma_value_().value()

    def trace(cls):
        return cls._get_magma_value_().trace()

    def driven(cls):
        return cls._get_magma_value_().driven()

    @property
    def name(cls):
        return cls._get_magma_value_().name



from .compatibility import IntegerTypes
from abc import abstractmethod


__all__ = ['AnonRef', 'InstRef', 'DefnRef', 'ArrayRef', 'TupleRef']
__all__ += ['LazyDefnRef', 'LazyCircuit']


class Ref:
    @abstractmethod
    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return self.qualifiedname()

    @abstractmethod
    def qualifiedname(self, sep="."):
        raise NotImplementedError()

    @abstractmethod
    def anon(self):
        raise NotImplementedError()

    def verilog_name(self):
        return self.qualifiedname("_")


class AnonRef(Ref):
    def __init__(self):
        self.name = None

    def __str__(self):
        return f"AnonymousValue_{id(self)}"

    def qualifiedname(self, sep='.'):
        return f"AnonymousValue_{id(self)}"

    def anon(self):
        return True


class NamedRef(Ref):
    def __init__(self, name):
        if not isinstance(name, (str, int)):
            raise TypeError("Expected string or int")
        self.name = name

    def __str__(self):
        return self.name

    def qualifiedname(self, sep="."):
        return self.name

    def anon(self):
        return False


class InstRef(NamedRef):
    def __init__(self, inst, name):
        super().__init__(name)
        if not inst:
            raise ValueError(f"Bad inst: {inst}")
        self.inst = inst

    def qualifiedname(self, sep="."):
        name = self.name
        if isinstance(self.name, IntegerTypes):
            # Hack, Hack, Hack!
            # NOTE: This is used for verilog instances that don't use named
            # port (wired by index instead), so the ports are referred to by
            # index instead of name and we use the array indexing syntax to
            # represent them
            # See mantle's generic verilog target for example use case
            if sep == ".":
                return f"{self.inst.name}[{self.name}]"
        return self.inst.name + sep + str(name)


class DefnRef(NamedRef):
    def __init__(self, defn, name):
        super().__init__(name)
        if not defn:
            raise ValueError(f"Bad defn: {defn}")
        self.defn = defn

    def qualifiedname(self, sep="."):
        if sep == ".":
            return self.defn.__name__ + sep + self.name
        return self.name


class LazyCircuit:
    name = ""


class LazyDefnRef(DefnRef):
    def __init__(self, name):
        self.name = name
        self._defn = None

    @property
    def defn(self):
        if self._defn is not None:
            return self._defn
        return LazyCircuit

    def qualifiedname(self, sep="."):
        return super().qualifiedname(sep)

    def set_defn(self, defn):
        if self._defn is not None:
            raise Exception("Can only set definition of LazyDefnRef once")
        self._defn = defn


class ArrayRef(Ref):
   def __init__(self, array, index):
       self.array = array
       self.index = index

   def __str__(self):
       return self.qualifiedname()

   def qualifiedname(self, sep="."):
       return f"{self.array.name.qualifiedname(sep=sep)}[{self.index}]"

   def anon(self):
       return self.array.name.anon()


class TupleRef(Ref):
    def __init__(self, tuple, index):
        self.tuple = tuple
        self.index = index

    def __str__(self):
        return self.qualifiedname()

    def qualifiedname(self, sep="."):
        try:
            int(self.index)
            return self.tuple.name.qualifiedname(sep=sep) + "[" + str(self.index) + "]"
        except ValueError:
            return self.tuple.name.qualifiedname(sep=sep) + sep + str(self.index)

    def verilog_name(self):
        return self.tuple.name.verilog_name() + "_" + str(self.index)

    def anon(self):
        return self.tuple.name.anon()



from abc import abstractmethod
import functools
import collections


class ParamDict(dict):
    """
    Hashable dictionary for simple key: value parameters
    """
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class GeneratorMeta(type):
    def __new__(mcs, name, bases, attrs):
        attrs["bind_generators"] = []
        cls = super().__new__(mcs, name, bases, attrs)
        if cls.cache:
            cls.generate = functools.lru_cache(maxsize=None)(cls.generate)

        old_generate = cls.generate

        def generate_wrapper(*args, **kwargs):
            result = old_generate(*args, **kwargs)
            for gen in cls.bind_generators:
                gen.generate_bind(result, *args, **kwargs)
            return result

        cls.generate = generate_wrapper
        return cls

    def __call__(cls, *args, name=None, **kwargs):
        """
        name is reserved kwarg for naming instances

        TODO: Generalize handling of instance parameters
        """
        inst_kwargs = {}
        if name is not None:
            inst_kwargs["name"] = name
        return cls.generate(*args, **kwargs)(**inst_kwargs)


class Generator(metaclass=GeneratorMeta):
    # User can disable cacheing by setting this attribute to False
    cache = True

    @staticmethod
    @abstractmethod
    def generate(*args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def bind(cls, monitor):
        cls.bind_generators.append(monitor)



import magma as m
from magma.view import InstView, PortView


def value_to_verilog_name(value):
    if isinstance(value, m.Array) and not issubclass(value.T, m.Digital):
        elems = ", ".join(value_to_verilog_name(t) for t in reversed(value))
        return f"'{{{elems}}}"
    elif isinstance(value, m.Tuple):
        raise NotImplementedError("Inlining unflattened tuple")
    elif isinstance(value, PortView):
        parent_name = value_to_verilog_name(value.parent)
        return f"{parent_name}.{verilog_name(value.port.name)}"
    elif isinstance(value, InstView):
        prefix = ""
        if isinstance(value.parent, m.Circuit):
            prefix = value.parent.name + "."
        else:
            prefix = value_to_verilog_name(value.parent) + "."
        return prefix + value.inst.name
    return verilog_name(value.name)


def verilog_name(name, inst_sep="."):
    if isinstance(name, m.ref.DefnRef):
        return str(name)
    if isinstance(name, m.ref.InstRef):
        return f"{name.inst.name}{inst_sep}{str(name)}"
    if isinstance(name, m.ref.NamedRef):
        return str(name)
    if isinstance(name, m.ref.ArrayRef):
        array_name = verilog_name(name.array.name)
        return f"{array_name}_{name.index}"
    if isinstance(name, m.ref.TupleRef):
        tuple_name = verilog_name(name.tuple.name)
        index = name.index
        try:
            int(index)
            # python/coreir don't allow pure integer names
            index = f"_{index}"
        except ValueError:
            pass
        return f"{tuple_name}_{index}"
    raise NotImplementedError(name, type(name))



from functools import lru_cache
import weakref
import magma as m
from abc import ABCMeta
from .t import Kind, Direction, Type
from .debug import debug_wire, get_callee_frame_info
from .compatibility import IntegerTypes
from .logging import root_logger
from .wire import Wire


_logger = root_logger()


class DigitalMeta(ABCMeta, Kind):
    _class_cache = weakref.WeakValueDictionary()

    def __new__(cls, name, bases, namespace, info=(None, None), **kwargs):
        # TODO: A lot of this code is shared with AbstractBitVectorMeta, we
        # should refactor to reuse
        if '_info_' in namespace:
            raise TypeError(
                'class attribute _info_ is reversed by the type machinery')

        direction = info[1]
        for base in bases:
            if getattr(base, 'is_directed', False):
                if direction is None:
                    direction = base.direction
                elif direction != base.direction:
                    raise TypeError(
                        "Can't inherit from multiple different directions")

        namespace['_info_'] = info[0], direction
        type_ = super().__new__(cls, name, bases, namespace, **kwargs)
        if direction is None:
            # class is unundirected so t.unundirected_t -> t
            type_._info_ = type_, direction
        elif info[0] is None:
            # class inherited from directed types so there is no unundirected_t
            type_._info_ = None, direction

        return type_

    def __getitem__(cls, direction: Direction) -> 'DigitalMeta':
        mcs = type(cls)
        try:
            return mcs._class_cache[cls, direction]
        except KeyError:
            pass

        if not isinstance(direction, Direction):
            raise TypeError('Direction of Digital must be an instance of '
                            'm.Direction')

        if cls.is_directed:
            if direction == direction.Undirected:
                return cls.undirected_t
            if direction == cls.direction:
                return cls
            else:
                return cls.undirected_t[direction]
        if direction == direction.Undirected:
            return cls

        bases = [cls]
        bases.extend(b[direction] for b in cls.__bases__ if isinstance(b, mcs))
        bases = tuple(bases)
        orig_name = cls.__name__
        class_name = '{}[{}]'.format(orig_name, direction.name)
        type_ = mcs(class_name, bases, {"orig_name": orig_name},
                    info=(cls, direction))
        type_.__module__ = cls.__module__
        mcs._class_cache[cls, direction] = type_
        return type_

    def __call__(cls, value=None, *args, **kwargs):
        if value is not None:
            if isinstance(value, (bool, IntegerTypes)):
                return m.VCC if value else m.GND
        result = super().__call__(*args, **kwargs)
        if value is not None:
            assert isinstance(value, Digital), type(value)
            result(value)
        return result

    def __repr__(cls):
        return str(cls)

    def __str__(cls):
        name = getattr(cls, "orig_name", cls.__name__)
        if cls.is_input():
            name = f"In({name})"
        elif cls.is_output():
            name = f"Out({name})"
        elif cls.is_inout():
            name = f"InOut({name})"
        return name

    @property
    def undirected_t(cls) -> 'DigitalMeta':
        t = cls._info_[0]
        if t is not None:
            return t
        else:
            raise AttributeError('type {} has no undirected_t'.format(cls))

    @property
    def direction(cls) -> int:
        return cls._info_[1]

    @property
    def is_directed(cls) -> bool:
        return cls.direction is not None

    def __len__(cls):
        return 1

    def qualify(cls, direction):
        return cls[direction]

    def __eq__(cls, rhs):
        return cls is rhs

    __hash__ = type.__hash__


class Digital(Type, metaclass=DigitalMeta):
    def __init__(self, value=None, name=None):
        super().__init__(name=name)
        self._wire = Wire(self)

    @classmethod
    def is_oriented(cls, direction):
        return cls.direction == direction

    def __call__(self, output):
        return self.wire(output, get_callee_frame_info())

    @debug_wire
    def wire(self, o, debug_info):
        i = self
        # promote integer types to LOW/HIGH
        if isinstance(o, IntegerTypes):
            o = m.HIGH if o else m.LOW

        if not isinstance(o, Digital):
            _logger.error(f'Cannot wire {i.debug_name} (type={type(i)}) to {o} '
                          f'(type={type(o)}) because {o.debug_name} is not a '
                          f'Digital', debug_info=debug_info)
            return

        i._wire.connect(o._wire, debug_info)
        i.debug_info = debug_info
        o.debug_info = debug_info

    def wired(self):
        return self._wire.wired()

    # return the input or output Bit connected to this Bit
    def trace(self):
        return self._wire.trace()

    # return the output Bit connected to this input Bit
    def value(self):
        return self._wire.value()

    def driven(self):
        return self._wire.driven()

    def flatten(self):
        return [self]

    def const(self):
        return self is m.VCC or self is m.GND

    def unwire(i, o):
        i._wire.unwire(o._wire)

    @classmethod
    def flat_length(cls):
        return 1

    # Loam methods, TODO: rework this interface
    def getinst(self):
        t = self.trace()
        return t.name.inst if t is not None else None

    def getgpio(self):
        return self.getinst()

    def unused(self):
        if self.is_input() or self.is_inout():
            raise TypeError("unused cannot be used with input/inout")
        m.wire(self, DefineUnused()().I)

    def undriven(self):
        if self.is_output() or self.is_inout():
            raise TypeError("undriven cannot be used with output/inout")
        m.wire(DefineUndriven()().O, self)

    def as_bits(self):
        result = m.Bits[1]()
        result[0] @= self
        return result

    @classmethod
    def from_bits(cls, other):
        if not isinstance(other, m.Bits) and not len(other) == 1:
            raise TypeError("Can only convert from Bits[1] to Bit")
        return other[0]

    @classmethod
    def is_mixed(cls):
        return False


def make_Define(_name, port, direction):
    @lru_cache(maxsize=None)
    def DefineCorebit():
        class _Primitive(m.Circuit):
            renamed_ports = m.circuit.coreir_port_mapping
            name = f"corebit_{_name}"
            coreir_name = _name
            coreir_lib = "corebit"

            def simulate(self, value_store, state_store):
                pass

            io = m.IO(**{port: direction(Digital)})
        return _Primitive
    return DefineCorebit


DefineUndriven = make_Define("undriven", "O", m.Out)
DefineUnused = make_Define("term", "I", m.In)
